@@ -15,8 +15,10 @@ use Data::Dumper qw/Dumper/;
 use English qw/ -no_match_vars /;
 use base qw/Exporter/;
 use Path::Tiny;
+use IO::Handle;
+use POSIX qw/:errno_h/;
 
-our $VERSION     = '0.16';
+our $VERSION     = '0.17';
 our @EXPORT_OK   = qw/hosts_from_map is_host multi_run shell_quote tmux/;
 our %EXPORT_TAGS = ();
 
@@ -113,7 +115,18 @@ sub multi_run {
             elsif ( defined $child ) {
                 # child code
                 if ( $option->{interleave} ) {
-                    exec "$cmd 2>&1";
+                    require IPC::Open3::Callback;
+                    my ($pid, $in, $out, $err) = IPC::Open3::Callback::safe_open3($cmd);
+
+                    close $in;
+                    $out->blocking(0);
+                    $err->blocking(0);
+                    while ($out && $err) {
+                        $out = _read_label_line($out, \*STDOUT, $host);
+                        $err = _read_label_line($err, \*STDERR, $host);
+                    }
+                    waitpid $pid, 0;
+                    exit 0;
                 }
                 else {
                     my $out = `$cmd 2>&1`;
@@ -136,6 +149,35 @@ sub multi_run {
     wait;
 }
 
+sub _read_label_line {
+    my ($in_fh, $out_fh, $host) = @_;
+    state %hosts;
+    my @colours = (qw/
+        red     on_red     bright_red
+        green   on_green   bright_green
+        blue    on_blue    bright_blue
+        magenta on_magenta bright_magenta
+        cyan    on_cyan
+        yellow  on_yellow
+    /);
+    return if !$in_fh;
+
+    my $line = <$in_fh>;
+
+    if ( !defined $line && $! != EAGAIN ) {
+        close $in_fh;
+        return;
+    }
+
+    if (defined $line) {
+        $hosts{$host} ||= $colours[rand @colours];
+        require Term::ANSIColor;
+        print {$out_fh} '[', Term::ANSIColor::colored($host, $hosts{$host}), '] ', $line;
+    }
+
+    return $in_fh;
+}
+
 sub tmux {
     my (@commands) = @_;
 
@@ -224,7 +266,7 @@ App::MultiSsh - Multi host ssh executer
 
 =head1 VERSION
 
-This documentation refers to App::MultiSsh version 0.16
+This documentation refers to App::MultiSsh version 0.17
 
 =head1 SYNOPSIS
 
